#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from osc_lib import exceptions

from heat.common import template_format
from heat.engine.resources.openstack.octavia import health_monitor
from heat.tests import common
from heat.tests.openstack.octavia import inline_templates
from heat.tests import utils


class HealthMonitorTest(common.HeatTestCase):

    def _create_stack(self, tmpl=inline_templates.MONITOR_TEMPLATE):
        self.t = template_format.parse(tmpl)
        self.stack = utils.parse_stack(self.t)
        self.healthmonitor = self.stack['monitor']

        self.octavia_client = mock.MagicMock()
        self.healthmonitor.client = mock.MagicMock(
            return_value=self.octavia_client)
        self.healthmonitor.client_plugin().client = mock.MagicMock(
            return_value=self.octavia_client)

    def test_resource_mapping(self):
        mapping = health_monitor.resource_mapping()
        self.assertEqual(health_monitor.HealthMonitor,
                         mapping['OS::Octavia::HealthMonitor'])

    def test_create(self):
        self._create_stack()
        self.octavia_client.health_monitor_show.side_effect = [
            {'provisioning_status': 'PENDING_CREATE'},
            {'provisioning_status': 'PENDING_CREATE'},
            {'provisioning_status': 'ACTIVE'},
        ]
        self.octavia_client.health_monitor_create.side_effect = [
            exceptions.Conflict(409), {'healthmonitor': {'id': '1234'}}
        ]
        expected = {
            'healthmonitor': {
                'admin_state_up': True,
                'delay': 3,
                'expected_codes': '200-202',
                'http_method': 'HEAD',
                'max_retries': 5,
                'pool_id': '123',
                'timeout': 10,
                'type': 'HTTP',
                'url_path': '/health'
            }
        }

        props = self.healthmonitor.handle_create()

        self.assertFalse(self.healthmonitor.check_create_complete(props))
        self.octavia_client.health_monitor_create.assert_called_with(
            json=expected)
        self.assertFalse(self.healthmonitor.check_create_complete(props))
        self.octavia_client.health_monitor_create.assert_called_with(
            json=expected)
        self.assertFalse(self.healthmonitor.check_create_complete(props))
        self.assertTrue(self.healthmonitor.check_create_complete(props))

    def test_show_resource(self):
        self._create_stack()
        self.healthmonitor.resource_id_set('1234')

        self.assertTrue(self.healthmonitor._show_resource())

        self.octavia_client.health_monitor_show.assert_called_with(
            '1234')

    def test_update(self):
        self._create_stack()
        self.healthmonitor.resource_id_set('1234')
        self.octavia_client.health_monitor_show.side_effect = [
            {'provisioning_status': 'PENDING_UPDATE'},
            {'provisioning_status': 'PENDING_UPDATE'},
            {'provisioning_status': 'ACTIVE'},
        ]
        self.octavia_client.health_monitor_set.side_effect = [
            exceptions.Conflict(409), None]
        prop_diff = {
            'admin_state_up': False,
        }

        prop_diff = self.healthmonitor.handle_update(None, None, prop_diff)

        self.assertFalse(self.healthmonitor.check_update_complete(prop_diff))
        self.assertFalse(self.healthmonitor._update_called)
        self.octavia_client.health_monitor_set.assert_called_with(
            '1234', json={'healthmonitor': prop_diff})
        self.assertFalse(self.healthmonitor.check_update_complete(prop_diff))
        self.assertTrue(self.healthmonitor._update_called)
        self.octavia_client.health_monitor_set.assert_called_with(
            '1234', json={'healthmonitor': prop_diff})
        self.assertFalse(self.healthmonitor.check_update_complete(prop_diff))
        self.assertTrue(self.healthmonitor.check_update_complete(prop_diff))

    def test_delete(self):
        self._create_stack()
        self.healthmonitor.resource_id_set('1234')
        self.octavia_client.health_monitor_show.side_effect = [
            {'provisioning_status': 'PENDING_DELETE'},
            {'provisioning_status': 'PENDING_DELETE'},
            {'provisioning_status': 'DELETED'},
        ]
        self.octavia_client.health_monitor_delete.side_effect = [
            exceptions.Conflict(409),
            None]

        self.healthmonitor.handle_delete()

        self.assertFalse(self.healthmonitor.check_delete_complete(None))
        self.assertFalse(self.healthmonitor._delete_called)
        self.octavia_client.health_monitor_delete.assert_called_with(
            '1234')
        self.assertFalse(self.healthmonitor.check_delete_complete(None))
        self.assertTrue(self.healthmonitor._delete_called)
        self.octavia_client.health_monitor_delete.assert_called_with(
            '1234')
        self.assertTrue(self.healthmonitor.check_delete_complete(None))

    def test_delete_failed(self):
        self._create_stack()
        self.healthmonitor.resource_id_set('1234')
        self.octavia_client.health_monitor_delete.side_effect = (
            exceptions.Unauthorized(401))

        self.healthmonitor.handle_delete()
        self.assertRaises(exceptions.Unauthorized,
                          self.healthmonitor.check_delete_complete, None)

        self.octavia_client.health_monitor_delete.assert_called_with(
            '1234')



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Add extra_data column to sync_point table

Revision ID: d2db381e3324
Revises: c6214ca60943
Create Date: 2025-01-03

"""
from alembic import op
import sqlalchemy as sa

from heat.db import types

# revision identifiers, used by Alembic.
revision = 'd2db381e3324'
down_revision = 'c6214ca60943'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('sync_point',
                  sa.Column('extra_data', types.Json, nullable=True))


def downgrade():
    op.drop_column('sync_point', 'extra_data')



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""make snapshot state aware

Revision ID: 97b2a986f922
Revises: 88be64598c0f
Create Date: 2026-02-15 14:09:59.134120
"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '97b2a986f922'
down_revision = '88be64598c0f'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('snapshot',
                  sa.Column('action', sa.String(255), nullable=True))

    connection = op.get_bind()
    snapshot_table = sa.Table('snapshot', sa.MetaData(),
                              autoload_with=connection)
    op.execute(
        snapshot_table.update().where(
            snapshot_table.c.action.is_(None)
        ).values({
            'action': 'CREATE'
        })
    )


def downgrade():
    op.drop_column('snapshot', 'action')



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import tenacity

from oslo_log import log as logging

from heat.common import exception
from heat.objects import sync_point as sync_point_object

LOG = logging.getLogger(__name__)


KEY_SEPERATOR = ':'


def _dump_list(items, separator=', '):
    return separator.join(map(str, items))


def make_key(*components):
    assert len(components) >= 2
    return _dump_list(components, KEY_SEPERATOR)


def create(context, entity_id, traversal_id, is_update, stack_id):
    """Creates a sync point entry in DB."""
    values = {'entity_id': entity_id, 'traversal_id': traversal_id,
              'is_update': is_update, 'atomic_key': 0,
              'stack_id': stack_id, 'input_data': {}}
    return sync_point_object.SyncPoint.create(context, values)


def get(context, entity_id, traversal_id, is_update):
    """Retrieves a sync point entry from DB."""
    sync_point = sync_point_object.SyncPoint.get_by_key(context, entity_id,
                                                        traversal_id,
                                                        is_update)
    if sync_point is None:
        key = (entity_id, traversal_id, is_update)
        raise exception.EntityNotFound(entity='Sync Point', name=key)

    return sync_point


def delete_all(context, stack_id, traversal_id):
    """Deletes all sync points of a stack associated with a traversal_id."""
    return sync_point_object.SyncPoint.delete_all_by_stack_and_traversal(
        context, stack_id, traversal_id
    )


def update_input_data(context, entity_id, current_traversal,
                      is_update, atomic_key, input_data=None, extra_data=None):
    rows_updated = sync_point_object.SyncPoint.update_input_data(
        context, entity_id, current_traversal, is_update, atomic_key,
        input_data, extra_data)

    return rows_updated


def str_pack_tuple(t):
    return 'tuple:' + str(tuple(t))


def _str_unpack_tuple(s):
    s = s[s.index(':') + 1:]
    return ast.literal_eval(s)


def _deserialize(d):
    d2 = {}
    for k, v in d.items():
        if isinstance(k, str) and k.startswith('tuple:('):
            k = _str_unpack_tuple(k)
        if isinstance(v, dict):
            v = _deserialize(v)
        d2[k] = v
    return d2


def _serialize(d):
    d2 = {}
    for k, v in d.items():
        if isinstance(k, tuple):
            k = str_pack_tuple(k)
        if isinstance(v, dict):
            v = _serialize(v)
        d2[k] = v
    return d2


def deserialize_input_data(db_input_data):
    return _deserialize_data(db_input_data, 'input_data')


def deserialize_extra_data(db_extra_data):
    return _deserialize_data(db_extra_data, 'extra_data')


def _deserialize_data(db_data, key):
    db_data = db_data.get(key)
    if not db_data:
        return {}
    return dict(_deserialize(db_data))


def serialize_input_data(input_data):
    return _serialize_data(input_data, 'input_data')


def serialize_extra_data(extra_data):
    return _serialize_data(extra_data, 'extra_data')


def _serialize_data(data, key):
    return {key: _serialize(data)}


def update_sync_point(cnxt, entity_id, current_traversal, is_update,
                      predecessors, new_data, new_resource_failures=None,
                      is_skip=False):
    """Update a sync point atomically with new data and failures.

    Retry waits up to 60 seconds at most, with exponentially increasing
    amounts of jitter per resource still outstanding.
    """
    wait_strategy = tenacity.wait_random_exponential(max=60)

    def init_jitter(existing_input_data):
        nconflicts = max(0, len(predecessors) - len(existing_input_data) - 1)
        # 10ms per potential conflict, up to a max of 10s in total
        return min(nconflicts, 1000) * 0.01

    @tenacity.retry(
        retry=tenacity.retry_if_result(lambda r: r is None),
        wait=wait_strategy
    )
    def _sync():
        sync_point = get(cnxt, entity_id, current_traversal, is_update)
        input_data = deserialize_input_data(sync_point.input_data)
        extra_data = deserialize_extra_data(
            sync_point.extra_data) if sync_point.extra_data is not None else {}
        resource_failures = extra_data.get("resource_failures", {})
        skip_propagate = extra_data.get("skip_propagate", False)
        if new_resource_failures is not None:
            resource_failures.update(new_resource_failures)
            extra_data.update({"resource_failures": resource_failures})
        if is_skip:
            extra_data.update({"skip_propagate": is_skip})
            skip_propagate = is_skip
        if not extra_data:
            extra_data = None
        else:
            extra_data = serialize_extra_data(extra_data)
        wait_strategy.multiplier = init_jitter(input_data)
        if new_data is not None:
            input_data.update(new_data)
        rows_updated = update_input_data(
            cnxt, entity_id, current_traversal, is_update,
            sync_point.atomic_key,
            serialize_input_data(input_data), extra_data)
        return (input_data, resource_failures,
                skip_propagate) if rows_updated else None
    return _sync()


def sync(cnxt, entity_id, current_traversal, is_update, propagate,
         predecessors, new_data, new_resource_failures=None,
         is_skip=False):
    """Synchronize resource state and propagate when all predecessors done.

    This function updates the sync point with new data and resource failures,
    and calls the propagate callback when all predecessors have reported.
    """
    result = update_sync_point(
        cnxt, entity_id, current_traversal, is_update,
        predecessors, new_data, new_resource_failures,
        is_skip=is_skip)
    if result is None:
        # Sync point update failed (possibly deleted by another traversal)
        LOG.warning('[%s] Sync point update failed for entity %s',
                    current_traversal, entity_id)
        return
    input_data, resource_failures, skip_propagate = result
    waiting = predecessors - set(input_data)
    key = make_key(entity_id, current_traversal, is_update)
    if waiting:
        LOG.debug('[%s] Waiting %s: Got %s; still need %s',
                  key, entity_id, _dump_list(input_data), _dump_list(waiting))
    else:
        LOG.debug('[%s] Ready %s: Got %s',
                  key, entity_id, _dump_list(input_data))
        propagate(entity_id, serialize_input_data(input_data),
                  resource_failures, skip_propagate)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Add resource_snapshot table

Revision ID: 88be64598c0f
Revises: d2db381e3324
Create Date: 2025-01-24

"""
from alembic import op
import sqlalchemy as sa

from heat.db import types

# revision identifiers, used by Alembic.
revision = '88be64598c0f'
down_revision = 'd2db381e3324'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'resource_snapshot',
        sa.Column('id', sa.String(36), primary_key=True, nullable=False),
        sa.Column('snapshot_id', sa.String(36),
                  sa.ForeignKey('snapshot.id'), nullable=False),
        sa.Column('resource_name', sa.String(255), nullable=False),
        sa.Column('created_at', sa.DateTime),
        sa.Column('updated_at', sa.DateTime),
        sa.Column('data', types.Json),
        sa.Index('ix_resource_snapshot_snapshot_id', 'snapshot_id'),
        mysql_engine='InnoDB',
    )


def downgrade():
    op.drop_table('resource_snapshot')



#
# Copyright 2010-2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from unittest import mock

import json
from oslo_config import cfg
import webob

from heat.api.aws import exception as aws_exception
from heat.common import exception
from heat.common import wsgi
from heat.tests import common


class ResourceTest(common.HeatTestCase):

    def test_get_action_args(self):
        env = {
            'wsgiorg.routing_args': [
                None,
                {
                    'controller': None,
                    'format': None,
                    'action': 'update',
                    'id': 12,
                },
            ],
        }

        expected = {'action': 'update', 'id': 12}
        actual = wsgi.Resource(None, None, None).get_action_args(env)

        self.assertEqual(expected, actual)

    def test_get_action_args_invalid_index(self):
        env = {'wsgiorg.routing_args': []}
        expected = {}
        actual = wsgi.Resource(None, None, None).get_action_args(env)
        self.assertEqual(expected, actual)

    def test_get_action_args_del_controller_error(self):
        actions = {'format': None,
                   'action': 'update',
                   'id': 12}
        env = {'wsgiorg.routing_args': [None, actions]}
        expected = {'action': 'update', 'id': 12}
        actual = wsgi.Resource(None, None, None).get_action_args(env)
        self.assertEqual(expected, actual)

    def test_get_action_args_del_format_error(self):
        actions = {'action': 'update', 'id': 12}
        env = {'wsgiorg.routing_args': [None, actions]}
        expected = {'action': 'update', 'id': 12}
        actual = wsgi.Resource(None, None, None).get_action_args(env)
        self.assertEqual(expected, actual)

    def test_dispatch(self):
        class Controller(object):
            def index(self, shirt, pants=None):
                return (shirt, pants)

        resource = wsgi.Resource(None, None, None)
        actual = resource.dispatch(Controller(), 'index', 'on', pants='off')
        expected = ('on', 'off')
        self.assertEqual(expected, actual)

    def test_dispatch_default(self):
        class Controller(object):
            def default(self, shirt, pants=None):
                return (shirt, pants)

        resource = wsgi.Resource(None, None, None)
        actual = resource.dispatch(Controller(), 'index', 'on', pants='off')
        expected = ('on', 'off')
        self.assertEqual(expected, actual)

    def test_dispatch_no_default(self):
        class Controller(object):
            def show(self, shirt, pants=None):
                return (shirt, pants)

        resource = wsgi.Resource(None, None, None)
        self.assertRaises(AttributeError, resource.dispatch, Controller(),
                          'index', 'on', pants='off')

    def test_resource_call_error_handle(self):
        class Controller(object):
            def delete(self, req, identity):
                return (req, identity)

        actions = {'action': 'delete', 'id': 12, 'body': 'data'}
        env = {'wsgiorg.routing_args': [None, actions]}
        request = wsgi.Request.blank('/tests/123', environ=env)
        request.body = b'{"foo" : "value"}'
        resource = wsgi.Resource(Controller(),
                                 wsgi.JSONRequestDeserializer(),
                                 None)
        # The Resource does not throw webob.HTTPExceptions, since they
        # would be considered responses by wsgi and the request flow would end,
        # instead they are wrapped so they can reach the fault application
        # where they are converted to a nice JSON/XML response
        e = self.assertRaises(exception.HTTPExceptionDisguise,
                              resource, request)
        self.assertIsInstance(e.exc, webob.exc.HTTPBadRequest)

    @mock.patch.object(wsgi, 'translate_exception')
    def test_resource_call_error_handle_localized(self, mock_translate):
        class Controller(object):
            def delete(self, req, identity):
                return (req, identity)

        def fake_translate_exception(ex, locale):
            return translated_ex

        mock_translate.side_effect = fake_translate_exception
        actions = {'action': 'delete', 'id': 12, 'body': 'data'}
        env = {'wsgiorg.routing_args': [None, actions]}
        request = wsgi.Request.blank('/tests/123', environ=env)
        request.body = b'{"foo" : "value"}'
        message_es = "No Encontrado"
        translated_ex = webob.exc.HTTPBadRequest(message_es)

        resource = wsgi.Resource(Controller(),
                                 wsgi.JSONRequestDeserializer(),
                                 None)
        e = self.assertRaises(exception.HTTPExceptionDisguise,
                              resource, request)
        self.assertEqual(message_es, str(e.exc))


class ResourceExceptionHandlingTest(common.HeatTestCase):
    scenarios = [
        ('client_exceptions', dict(
            exception=exception.StackResourceLimitExceeded,
            exception_catch=exception.StackResourceLimitExceeded)),
        ('aws_exception', dict(
            exception=aws_exception.HeatAccessDeniedError,
            exception_catch=aws_exception.HeatAccessDeniedError)),
        ('webob_bad_request', dict(
            exception=webob.exc.HTTPBadRequest,
            exception_catch=exception.HTTPExceptionDisguise)),
        ('webob_not_found', dict(
            exception=webob.exc.HTTPNotFound,
            exception_catch=exception.HTTPExceptionDisguise)),
    ]

    def test_resource_client_exceptions_dont_log_error(self):
        class Controller(object):
            def __init__(self, exception_to_raise):
                self.exception_to_raise = exception_to_raise

            def raise_exception(self, req, body):
                raise self.exception_to_raise()

        actions = {'action': 'raise_exception', 'body': 'data'}
        env = {'wsgiorg.routing_args': [None, actions]}
        request = wsgi.Request.blank('/tests/123', environ=env)
        request.body = b'{"foo" : "value"}'
        resource = wsgi.Resource(Controller(self.exception),
                                 wsgi.JSONRequestDeserializer(),
                                 None)
        e = self.assertRaises(self.exception_catch, resource, request)
        e = e.exc if hasattr(e, 'exc') else e
        self.assertNotIn(str(e), self.LOG.output)


class JSONRequestDeserializerTest(common.HeatTestCase):

    def test_has_body_no_content_length(self):
        request = wsgi.Request.blank('/')
        request.method = 'POST'
        request.body = b'asdf'
        request.headers.pop('Content-Length')
        request.headers['Content-Type'] = 'application/json'
        self.assertFalse(wsgi.JSONRequestDeserializer().has_body(request))

    def test_has_body_zero_content_length(self):
        request = wsgi.Request.blank('/')
        request.method = 'POST'
        request.body = b'asdf'
        request.headers['Content-Length'] = 0
        request.headers['Content-Type'] = 'application/json'
        self.assertFalse(wsgi.JSONRequestDeserializer().has_body(request))

    def test_has_body_has_content_length_no_content_type(self):
        request = wsgi.Request.blank('/')
        request.method = 'POST'
        request.body = b'{"key": "value"}'
        self.assertIn('Content-Length', request.headers)
        self.assertTrue(wsgi.JSONRequestDeserializer().has_body(request))

    def test_has_body_has_content_length_plain_content_type(self):
        request = wsgi.Request.blank('/')
        request.method = 'POST'
        request.body = b'{"key": "value"}'
        self.assertIn('Content-Length', request.headers)
        request.headers['Content-Type'] = 'text/plain'
        self.assertTrue(wsgi.JSONRequestDeserializer().has_body(request))

    def test_has_body_has_content_type_malformed(self):
        request = wsgi.Request.blank('/')
        request.method = 'POST'
        request.body = b'asdf'
        self.assertIn('Content-Length', request.headers)
        request.headers['Content-Type'] = 'application/json'
        self.assertFalse(wsgi.JSONRequestDeserializer().has_body(request))

    def test_has_body_has_content_type(self):
        request = wsgi.Request.blank('/')
        request.method = 'POST'
        request.body = b'{"key": "value"}'
        self.assertIn('Content-Length', request.headers)
        request.headers['Content-Type'] = 'application/json'
        self.assertTrue(wsgi.JSONRequestDeserializer().has_body(request))

    def test_has_body_has_wrong_content_type(self):
        request = wsgi.Request.blank('/')
        request.method = 'POST'
        request.body = b'{"key": "value"}'
        self.assertIn('Content-Length', request.headers)
        request.headers['Content-Type'] = 'application/xml'
        self.assertFalse(wsgi.JSONRequestDeserializer().has_body(request))

    def test_has_body_has_aws_content_type_only(self):
        request = wsgi.Request.blank('/?ContentType=JSON')
        request.method = 'GET'
        request.body = b'{"key": "value"}'
        self.assertIn('Content-Length', request.headers)
        self.assertTrue(wsgi.JSONRequestDeserializer().has_body(request))

    def test_has_body_respect_aws_content_type(self):
        request = wsgi.Request.blank('/?ContentType=JSON')
        request.method = 'GET'
        request.body = b'{"key": "value"}'
        self.assertIn('Content-Length', request.headers)
        request.headers['Content-Type'] = 'application/xml'
        self.assertTrue(wsgi.JSONRequestDeserializer().has_body(request))

    def test_has_body_content_type_with_get(self):
        request = wsgi.Request.blank('/')
        request.method = 'GET'
        request.body = b'{"key": "value"}'
        self.assertIn('Content-Length', request.headers)
        self.assertTrue(wsgi.JSONRequestDeserializer().has_body(request))

    def test_no_body_no_content_length(self):
        request = wsgi.Request.blank('/')
        self.assertFalse(wsgi.JSONRequestDeserializer().has_body(request))

    def test_from_json(self):
        fixture = '{"key": "value"}'
        expected = {"key": "value"}
        actual = wsgi.JSONRequestDeserializer().from_json(fixture)
        self.assertEqual(expected, actual)

    def test_from_json_malformed(self):
        fixture = 'kjasdklfjsklajf'
        self.assertRaises(webob.exc.HTTPBadRequest,
                          wsgi.JSONRequestDeserializer().from_json, fixture)

    def test_default_no_body(self):
        request = wsgi.Request.blank('/')
        actual = wsgi.JSONRequestDeserializer().default(request)
        expected = {}
        self.assertEqual(expected, actual)

    def test_default_with_body(self):
        request = wsgi.Request.blank('/')
        request.method = 'POST'
        request.body = b'{"key": "value"}'
        actual = wsgi.JSONRequestDeserializer().default(request)
        expected = {"body": {"key": "value"}}
        self.assertEqual(expected, actual)

    def test_default_with_get_with_body(self):
        request = wsgi.Request.blank('/')
        request.method = 'GET'
        request.body = b'{"key": "value"}'
        actual = wsgi.JSONRequestDeserializer().default(request)
        expected = {"body": {"key": "value"}}
        self.assertEqual(expected, actual)

    def test_default_with_get_with_body_with_aws(self):
        request = wsgi.Request.blank('/?ContentType=JSON')
        request.method = 'GET'
        request.body = b'{"key": "value"}'
        actual = wsgi.JSONRequestDeserializer().default(request)
        expected = {"body": {"key": "value"}}
        self.assertEqual(expected, actual)

    def test_from_json_exceeds_max_json_mb(self):
        cfg.CONF.set_override('max_json_body_size', 10)
        body = json.dumps(['a'] * cfg.CONF.max_json_body_size)
        self.assertGreater(len(body), cfg.CONF.max_json_body_size)
        error = self.assertRaises(exception.RequestLimitExceeded,
                                  wsgi.JSONRequestDeserializer().from_json,
                                  body)
        msg = ('Request limit exceeded: JSON body size '
               '(%s bytes) exceeds maximum allowed size (%s bytes).' % (
                   len(body), cfg.CONF.max_json_body_size))
        self.assertEqual(msg, str(error))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import tenacity
import time

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils as oslo_timeutils
from oslo_utils import uuidutils

from heat.common import context as heat_context
from heat.common import exception
from heat.common import timeutils as heat_timeutils
from heat.engine import scheduler
from heat.engine import stack as parser
from heat.engine import sync_point
from heat.objects import resource_snapshot as rsrc_snapshot_objects
from heat.objects import snapshot as snapshot_object
from heat.rpc import worker_client as rpc_worker_client


LOG = logging.getLogger(__name__)


class Snapshot(object):
    """Operate snapshot actions under convergence."""

    ACTIONS = (
        CREATE, DELETE
    ) = (
        'CREATE', 'DELETE'
    )

    STATUSES = (
        IN_PROGRESS, FAILED, COMPLETE
    ) = (
        'IN_PROGRESS', 'FAILED', 'COMPLETE'
    )

    def __init__(self, context, snapshot_id, stack_id,
                 start_time, thread_group_mgr, resources=None, action=None,
                 timeout_mins=None, is_stack_delete=False,
                 current_traversal=None):
        self.context = context
        self.id = snapshot_id
        self.stack_id = stack_id
        self.action = self.DELETE if action is None else action
        self.is_stack_delete = is_stack_delete
        self.current_traversal = current_traversal if (
            current_traversal is not None) else uuidutils.generate_uuid()

        self.thread_group_mgr = thread_group_mgr
        self._worker_client = None

        self.start_time = start_time
        self.timeout_mins = timeout_mins
        self.resources = resources

    def delete_snapshot(self):
        snapshot_obj = snapshot_object.Snapshot.get_snapshot(
            self.context, self.id, load_rsrc_snapshot=True)
        # NOTE(tkajinam): Check None for old services without action field
        if (snapshot_obj.action in (None, self.CREATE) and
                snapshot_obj.status == self.IN_PROGRESS):
            msg = _('Deleting in-progress snapshot')
            raise exception.NotSupported(feature=msg)
        if self.resources is None:
            self.resources = snapshot_obj.data['resources']

        # TODO(ricolin) send notification and add event

        LOG.debug("Start deleting snapshot %s.", self.id)
        self.thread_group_mgr.start(self.stack_id, self.do_delete_snapshot)

    def do_delete_snapshot(self):
        if self.resources:
            sync_point.create(self.context, self.id,
                              self.current_traversal, True,
                              self.stack_id)
            LOG.debug("Create sync_point for (snapshot, traversal, "
                      "is_update): %s.",
                      (self.id, self.current_traversal, True))

            start_time = self.start_time.strftime(
                heat_timeutils.str_duration_format)

            for rsrc_name in self.resources:
                self.worker_client.check_resource_delete_snapshot(
                    self.context, self.id, rsrc_name, start_time,
                    is_stack_delete=self.is_stack_delete,
                    current_traversal=self.current_traversal)

            if scheduler.ENABLE_SLEEP:
                time.sleep(1)
        else:
            self.mark_complete()

    def mark_complete(self, predecessors=None, input_data=None):
        if predecessors is None:
            predecessors = []
        self.delete_snapshot_objs()
        # Clear sync_points for snapshot
        sync_point.delete_all(self.context, self.id, self.current_traversal)
        LOG.debug("Clear all sync_points for (snapshot, traversal): %s.",
                  (self.id, self.current_traversal))
        if self.is_stack_delete:
            if input_data is None:
                sender_key = parser.ConvergenceNode(self.id, True)
                input_data = {sender_key: None}
            try:
                update = sync_point.update_sync_point(
                    self.context, self.stack_id, self.current_traversal,
                    True, predecessors=predecessors, new_data=input_data)
                if not update:
                    LOG.debug("Fail to update stack %(stack)s sync_point entry"
                              "after snapshot %(snapshot_id)s competed",
                              {'snapshot_id': self.id, 'stack': self.stack_id})
                    return False
            except exception.EntityNotFound:
                LOG.debug("Ignore EntityNotFound: Stack sync_point entity %s "
                          "already deleted, this is because other snapshot "
                          "delete already failed.",
                          (self.stack_id, self.current_traversal, True))
        LOG.info("Snapshot %(snapshot_id)s for stack %(stack_id)s %(action)s "
                 "complete.",
                 {'snapshot_id': self.id, 'stack_id': self.stack_id,
                  'action': self.action})
        return True

    def mark_failed(self, rsrc_name, failure_reason):
        LOG.info("Snapshot %(snapshot_id)s for Stack %(stack_id)s %(action)s "
                 "failed. Snapshot resource: %(rsrc_name)s %(action)s failed "
                 "with reason: %(reason)s.",
                 {'snapshot_id': self.id, 'stack_id': self.stack_id,
                  'action': self.action, 'rsrc_name': rsrc_name,
                  'reason': failure_reason})
        try:
            snapshot_object.Snapshot.update(
                self.context, self.id,
                {'status': self.FAILED,
                 'action': self.action,
                 'status_reason': failure_reason})
        except Exception:
            LOG.debug("Failed to update snapshot %(snapshot_id)s status.",
                      {'snapshot_id': self.id})
        # Clear sync_points for snapshot
        sync_point.delete_all(self.context, self.id, self.current_traversal)
        LOG.debug("Clear all sync_points for (snapshot, traversal): %s.",
                  (self.id, self.current_traversal))
        if self.is_stack_delete:
            try:
                update = sync_point.update_sync_point(
                    self.context, self.stack_id, self.current_traversal,
                    True, predecessors=[], new_rsrc_failure=failure_reason)
                if not update:
                    LOG.debug("Fail to update stack %(stack)s sync_point "
                              "entry after snapshot %(snapshot_id)s failed.",
                              {'snapshot_id': self.id, 'stack': self.stack_id})
                    return False
            except exception.EntityNotFound:
                LOG.debug("Ignore EntityNotFound: Stack sync_point entity %s "
                          "already deleted, this is because other snapshot "
                          "delete already failed.",
                          (self.stack_id, self.current_traversal, True))
        return True

    def delete_snapshot_objs(self):
        LOG.debug("Start to clear objects for snapshot %(snapshot_id)s.",
                  {'snapshot_id': self.id})
        rsrc_snapshot_objects.ResourceSnapshot.delete_all_by_snapshot(
            self.context, self.id)
        snapshot_object.Snapshot.delete(self.context, self.id)

    @property
    def worker_client(self):
        """Return a client for making engine RPC calls."""
        if not self._worker_client:
            self._worker_client = rpc_worker_client.WorkerClient()
        return self._worker_client

    def timeout_secs(self):
        """Return the action timeout in seconds."""
        if self.timeout_mins is None:
            return cfg.CONF.stack_action_timeout

        return self.timeout_mins * 60

    def time_elapsed(self):
        """Time elapsed in seconds since the Snapshot operation started."""
        return (oslo_timeutils.utcnow() - self.start_time).total_seconds()

    def time_remaining(self):
        """Time left before Snapshot times out."""
        return self.timeout_secs() - self.time_elapsed()

    def has_timed_out(self):
        """Returns True if this Snapshot has timed-out."""
        return self.time_elapsed() > self.timeout_secs()


def delete_snapshots(context, snapshot_ids, stack_id, current_traversal,
                     start_time, thread_group_mgr, run_till_success=True):
    if not snapshot_ids:
        return
    # create sync_point entry for stack. When snapshot delete completed/failed,
    # it will update to sync_point entry for stack.
    sync_point.create(
        context, stack_id, current_traversal, True, stack_id)
    LOG.debug("Create sync_point for (stack, traversal, is_update): %s",
              (stack_id, current_traversal, True))

    # avoid reuse current db session to multiple threads.
    cnxt = context.to_dict()
    # Iterate over a copy to safely remove items
    for snapshot_id in list(snapshot_ids):
        snapshot = Snapshot(
            context=heat_context.RequestContext.from_dict(cnxt),
            snapshot_id=snapshot_id,
            stack_id=stack_id,
            start_time=start_time,
            thread_group_mgr=thread_group_mgr,
            action=Snapshot.DELETE,
            is_stack_delete=True,
            current_traversal=current_traversal)
        try:
            snapshot.delete_snapshot()
        except exception.NotFound:
            LOG.info("Snapshot %(snapshot)s for stack %(stack)s is "
                     "already deleted.",
                     {'snapshot': snapshot_id, 'stack': stack_id})
            # Since it's not found in object, we don't need to update it to
            # stack sync_point entry.
            snapshot_ids.remove(snapshot_id)
    if run_till_success:
        allow_running_time = cfg.CONF.stack_action_timeout - (
            oslo_timeutils.utcnow() - start_time).total_seconds()
        wait_strategy = tenacity.wait_random_exponential(max=60)
        count_num_snapshots = len(snapshot_ids)

        def init_jitter(existing_input_data):
            nconflicts = max(
                0, count_num_snapshots - len(existing_input_data) - 1)
            # 10ms per potential conflict, up to a max of 10s in total
            return min(nconflicts, 1000) * 0.01

        @tenacity.retry(
            retry=tenacity.retry_if_result(lambda r: r is False),
            wait=wait_strategy,
            stop=tenacity.stop_after_delay(allow_running_time)
        )
        def _wait_for_complete():
            s_p = sync_point.get(context, stack_id, current_traversal, True)
            input_data = sync_point.deserialize_input_data(s_p.input_data)
            extra_data = sync_point.deserialize_extra_data(
                s_p.extra_data) if s_p.extra_data is not None else {}
            rsrc_failures = extra_data.get("resource_failures", {})
            if rsrc_failures:
                return str(rsrc_failures)
            snaps = set(parser.ConvergenceNode(s, True) for s in snapshot_ids)
            waiting = snaps - set(input_data)
            wait_strategy.multiplier = init_jitter(input_data)
            if not waiting:
                return
            return False

        failure_reason = _wait_for_complete()
        # Clear sync_points for stack
        sync_point.delete_all(context, stack_id, current_traversal)
        LOG.debug("Clear all sync_points for (stack, traversal): %s.",
                  (stack_id, current_traversal))
        return failure_reason



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.engine.resources.openstack.heat import none_resource
from heat.engine import support


class MonascaNotification(none_resource.NoneResource):
    """Heat Template Resource for Monasca Notification.

    A resource which is used to notificate if there is some alarm.
    Monasca Notification helps to declare the hook points, which will be
    invoked once alarm is generated. This plugin helps to create, update and
    delete the notification.
    """

    support_status = support.SupportStatus(
        version='25.0.0',
        status=support.HIDDEN,
        message=_('Monasca project was retired'),
        previous_status=support.SupportStatus(
            version='22.0.0',
            status=support.DEPRECATED,
            message=_('Monasca project was marked inactive'),
            previous_status=support.SupportStatus(
                version='7.0.0',
                status=support.SUPPORTED,
                previous_status=support.SupportStatus(
                    version='5.0.0',
                    status=support.SUPPORTED
                ))))


def resource_mapping():
    return {
        'OS::Monasca::Notification': MonascaNotification
    }
